from typing import Union, List

from .chord import Chord


class ChordProgression:
    """ Class to handle chord progressions.

    Attributes:
        _chords: component chords of chord progression.
    """

    def __init__(self, initial_chords: Union[str, Chord, List[Union[str, Chord]]] = []):
        """ Constructor of ChordProgression instance.

        :param initial_chords: Initial chord or chords of the chord progressions
        """
        if isinstance(initial_chords, Chord):
            chords = [initial_chords]
        elif isinstance(initial_chords, str):
            chords = [self._as_chord(initial_chords)]
        elif isinstance(initial_chords, list):
            chords = [self._as_chord(chord) for chord in initial_chords]
        else:
            raise TypeError(
                f"Cannot initialize ChordProgression with argument of {type(initial_chords)} type")
        self._chords: List[Chord] = chords

    def __unicode__(self):
        return " | ".join([chord.chord for chord in self._chords])

    def __str__(self):
        return " | ".join([chord.chord for chord in self._chords])

    def __repr__(self):
        return f"<ChordProgression: {' | '.join([chord.chord for chord in self._chords])}>"

    def __add__(self, other):
        self._chords += other.chords
        return self

    def __len__(self):
        return len(self._chords)

    def __getitem__(self, item):
        return self._chords[item]

    def __setitem__(self, key, value):
        self._chords[key] = value

    def __eq__(self, other):
        if not isinstance(other, ChordProgression):
            raise TypeError(
                f"Cannot compare ChordProgression object with {type(other)} object")
        if len(self) != len(other):
            return False
        for c, o in zip(self, other):
            if c != o:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def chords(self) -> List[Chord]:
        """ Get component chords of chord progression """
        return self._chords

    def append(self, chord: Union[str, Chord]) -> None:
        """ Append a chord to chord progressions

        :param chord: A chord to append
        """
        self._chords.append(self._as_chord(chord))

    def insert(self, index: int, chord: Union[str, Chord]) -> None:
        """ Insert a chord to chord progressions

        :param index: Index to insert a chord
        :param chord: A chord to insert
        """
        self._chords.insert(index, self._as_chord(chord))

    def pop(self, index: int = -1) -> Chord:
        """ Pop a chord from chord progressions

        :param index: Index of the chord to pop (default: -1)
        """
        return self._chords.pop(index)

    def transpose(self, trans: int) -> None:
        """ Transpose whole chord progressions

        :param trans: Transpose key
        """
        for chord in self._chords:
            chord.transpose(trans)

    @staticmethod
    def _as_chord(chord: Union[str, Chord]) -> Chord:
        """ Convert from str to Chord instance if input is str

        :param chord: Chord name or Chord instance
        :return: Chord instance
        """
        if isinstance(chord, Chord):
            return chord
        elif isinstance(chord, str):
            return Chord(chord)
        else:
            raise TypeError("input type should be str or Chord instance.")



from typing import List, Union

from .constants import NOTE_VAL_DICT, VAL_NOTE_DICT
from .constants.scales import RELATIVE_KEY_DICT
from .parser import parse
from .quality import QualityManager, Quality
from .utils import transpose_note, display_appended, display_on, note_to_val, val_to_note


class Chord:
    """ Class to handle a chord.

    Attributes:
        _chord: Name of the chord. (e.g. C, Am7, F#m7-5/A)
        _root: The root note of chord. (e.g. C, A, F#)
        _quality: The quality of chord. (e.g. maj, m7, m7-5)
        _appended: The appended notes on chord.
        _on: The base note of slash chord.
    """

    def __init__(self, chord: str):
        """ Constructor of Chord instance

        :param chord: Name of chord (e.g. C, Am7, F#m7-5/A).
        """
        root, quality, appended, on = parse(chord)
        self._chord: str = chord
        self._root: str = root
        self._quality: Quality = quality
        self._appended: List[str] = appended
        self._on: str = on

        self._append_on_chord()

    def __unicode__(self):
        return self._chord

    def __str__(self):
        return self._chord

    def __repr__(self):
        return f"<Chord: {self._chord}>"

    def __eq__(self, other):
        if not isinstance(other, Chord):
            raise TypeError(f"Cannot compare Chord object with {type(other)} object")
        if note_to_val(self._root) != note_to_val(other.root):
            return False
        if self._quality != other.quality:
            return False
        if self._appended != other.appended:
            return False
        if self._on and other.on:
            if note_to_val(self._on) != note_to_val(other.on):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_note_index(
        cls,
        note: int,
        quality: str,
        scale: str,
        diatonic: bool = False,
        chromatic: int = 0,
    ) -> "Chord":
        """Create a Chord from note index in a scale

        Chord.from_note_index(1, "", "Cmaj") returns I of C major => Chord("C")
        Chord.from_note_index(3, "m7", "Fmaj") returns IIImin of F major => Chord("Am7")
        Chord.from_note_index(5, "7", "Amin") returns Vmin of A minor => Chord("E7")
        Chord.from_note_index(2, "", "Cmaj") returns II of C major => Chord("D")
        Chord.from_note_index(2, "m", "Cmaj") returns IImin of C major => Chord("Dm")
        Chord.from_note_index(2, "", "Cmaj", diatonic=True) returns IImin of C major => Chord("Dm")
        Chord.from_note_index(2, "", "Cmin", chromatic=-1) returns bII of C minor => Chord("Db")

        :param note: Scale degree of the chord's root (1-7)
        :param quality: Quality of the chord (e.g. m7, sus4)
        :param scale: Base scale (e.g. Cmaj, Amin, F#maj, Ebmin)
        :param diatonic: If True, chord quality is determined using the base scale (overrides :param quality)
        :param chromatic: Lower or raise the scale degree (and all notes of the chord) by semitone(s)
        """
        if not 1 <= note <= 8:
            raise ValueError(f"Invalid note {note}")
        relative_key = RELATIVE_KEY_DICT[scale[-3:]][note - 1]
        root_num = NOTE_VAL_DICT[scale[:-3]] + chromatic
        root = VAL_NOTE_DICT[(root_num + relative_key) % 12][0]

        scale_degrees = RELATIVE_KEY_DICT[scale[-3:]]

        if diatonic:
            # construct the chord based on scale degrees, within 1 octave
            third = scale_degrees[(note + 1) % 7]
            fifth = scale_degrees[(note + 3) % 7]
            seventh = scale_degrees[(note + 5) % 7]

            # adjust the chord to its root position (as a stack of thirds),
            # then set the root to 0
            # e.g. (9, 0, 4) -> [0, 3, 7]
            def get_diatonic_chord(chord):
                uninverted = []
                for note in chord:
                    if not uninverted:
                        uninverted.append(note)
                    elif note > uninverted[-1]:
                        uninverted.append(note)
                    else:
                        uninverted.append(note + 12)
                uninverted = [x - uninverted[0] for x in uninverted]
                return uninverted

            if quality in ["", "-", "maj", "m", "min"]:
                triad = (relative_key, third, fifth)
                q = get_diatonic_chord(triad)
            elif quality in ["7", "M7", "maj7", "m7"]:
                seventh_chord = (relative_key, third, fifth, seventh)
                q = get_diatonic_chord(seventh_chord)
            else:
                raise NotImplementedError("Only generic chords (triads, sevenths) are supported")

            # look up QualityManager to determine chord quality
            quality_manager = QualityManager()
            quality = quality_manager.find_quality_from_components(q)
            if not quality:
                raise RuntimeError(f"Quality with components {q} not found")

        return cls(f"{root}{quality}")

    @property
    def chord(self):
        """ The name of chord """
        return self._chord

    @property
    def root(self):
        """ The root note of chord """
        return self._root

    @property
    def quality(self):
        """ The quality of chord """
        return self._quality

    @property
    def appended(self):
        """ The appended notes on chord """
        return self._appended

    @property
    def on(self):
        """ The base note of slash chord """
        return self._on

    def info(self):
        """ Return information of chord to display """
        return f"""{self._chord}
root={self._root}
quality={self._quality}
appended={self._appended}
on={self._on}"""

    def transpose(self, trans: int, scale: str = "C") -> None:
        """ Transpose the chord

        :param trans: Transpose key
        :param scale: key scale
        """
        if not isinstance(trans, int):
            raise TypeError(f"Expected integers, not {type(trans)}")
        self._root = transpose_note(self._root, trans, scale)
        if self._on:
            self._on = transpose_note(self._on, trans, scale)
        self._reconfigure_chord()

    def components(self, visible: bool = True) -> Union[List[str], List[int]]:
        """ Return the component notes of chord

        :param visible: returns the name of notes if True else list of int
        :return: component notes of chord
        """
        return self._quality.get_components(root=self._root, visible=visible)

    def components_with_pitch(self, root_pitch: int) -> List[str]:
        """ Return the component notes of chord formatted like ["C4", "E4", "G4"]

        :param root_pitch: the pitch of the root note
        :return: component notes of chord
        """
        components = self._quality.get_components(root=self._root)
        if components[0] < 0:
            components = [c + 12 for c in components]
        return [f"{val_to_note(c, scale=self._root)}{root_pitch + c // 12}" for c in components]

    def _append_on_chord(self):
        if self._on:
            self._quality.append_on_chord(self.on, self.root)

    def _reconfigure_chord(self):
        # TODO: Use appended
        self._chord = "{}{}{}{}".format(self._root,
                                        self._quality.quality,
                                        display_appended(self._appended),
                                        display_on(self._on))



import unittest

from pychord.constants import NOTE_VAL_DICT, VAL_NOTE_DICT


class TestConstants(unittest.TestCase):

    def test_note_and_val(self):
        for note, val in NOTE_VAL_DICT.items():
            self.assertIn(note, VAL_NOTE_DICT[val])


if __name__ == '__main__':
    unittest.main()



from typing import List

from .chord import Chord
from .quality import QualityManager
from .utils import note_to_val


def find_chords_from_notes(notes: List[str]) -> List[Chord]:
    """ Find possible chords consisted from notes

    :param notes: List of note arranged from lower note. ex) ["C", "Eb", "G"]
    :return: List of chord
    """
    if not notes:
        raise ValueError("Please specify notes which consist a chord.")
    root = notes[0]
    root_and_positions = []
    for rotated_notes in get_all_rotated_notes(notes):
        rotated_root = rotated_notes[0]
        root_and_positions.append((rotated_root, notes_to_positions(rotated_notes, rotated_notes[0])))
    chords = []
    for temp_root, positions in root_and_positions:
        quality = QualityManager().find_quality_from_components(positions)
        if quality is None:
            continue
        if temp_root == root:
            chord = "{}{}".format(root, quality)
        else:
            chord = "{}{}/{}".format(temp_root, quality, root)
        chords.append(Chord(chord))
    return chords


def notes_to_positions(notes: List[str], root: str) -> List[int]:
    """ Get notes positions from the root note

    >>> notes_to_positions(["C", "E", "G"], "C")
    [0, 4, 7]

    :param notes: List of notes
    :param root: Root note
    :return: List of note positions
    """
    root_pos = note_to_val(root)
    current_pos = root_pos
    positions = []
    for note in notes:
        note_pos = note_to_val(note)
        if note_pos < current_pos:
            note_pos += 12 * ((current_pos - note_pos) // 12 + 1)
        positions.append(note_pos - root_pos)
        current_pos = note_pos
    return positions


def get_all_rotated_notes(notes: List[str]) -> List[List[str]]:
    """ Get all rotated notes

    get_all_rotated_notes([A,C,E]) -> [[A,C,E],[C,E,A],[E,A,C]]
    """
    notes_list = []
    for x in range(len(notes)):
        notes_list.append(notes[x:] + notes[:x])
    return notes_list



import unittest

from pychord import Chord


class TestChordComponent(unittest.TestCase):
    def _assert_components(self, chord, qualities, notes):
        """ Validates if a chord is made up of specified qualities and notes.

        :param str chord: A chord, specified as a string, e.g. "C7"
        :param qualities: The expected qualities of the chord, as a list of numbers
        :param notes: The expected notes of the chord, either as a list of strings,
          e.g. ["C", "E", "G", "Bb"] or a string, e.g. "C E G Bb"
        """
        c = Chord(chord)
        com0 = c.components(visible=False)
        self.assertEqual(com0, qualities)
        com1 = c.components(visible=True)
        if isinstance(notes, str):
            notes = notes.split()
        self.assertEqual(com1, notes)

    def test_normal_chord(self):
        self._assert_components("C", [0, 4, 7], ["C", "E", "G"])

    def test_minor_chord(self):
        self._assert_components("Am", [9, 12, 16], ["A", "C", "E"])

    def test_dim_chord(self):
        self._assert_components("Ddim", [2, 5, 8], ["D", "F", "G#"])

    def test_dim7_chord(self):
        self._assert_components("Cdim7", [0, 3, 6, 9], ["C", "Eb", "Gb", "A"])

    def test_aug_chord(self):
        self._assert_components("Eaug", [4, 8, 12], ["E", "G#", "C"])

    def test_slash_chord(self):
        self._assert_components("CM9/D", [-10, 0, 4, 7, 11], ["D", "C", "E", "G", "B"])

    def test_sus4_chord(self):
        self._assert_components("Fsus4", [5, 10, 12], ["F", "Bb", "C"])

    def test_seventh_chord(self):
        self._assert_components("G7", [7, 11, 14, 17], ["G", "B", "D", "F"])

    def test_sixth_chord(self):
        self._assert_components("C6", [0, 4, 7, 9], ["C", "E", "G", "A"])

    def test_appended_chord(self):
        self._assert_components("C#m7b9b5", [1, 4, 7, 11, 14], ["C#", "E", "G", "B", "D"])

    def test_major_add9(self):
        # major add 9 is a major chord with a Major ninth
        base = Chord("C")
        base0 = list(base.components(visible=False))
        base1 = list(base.components(visible=True))
        c = Chord("CMadd9")
        com0 = c.components(visible=False)
        self.assertEqual(com0, base0 + [14])
        com1 = c.components(visible=True)
        self.assertEqual(com1, base1 + ["D"])

    def test_5(self):
        self._assert_components("Db5", [1, 8], "Db Ab")

    def test_b5(self):
        self._assert_components("D(b5)", [2, 6, 8], "D F# G#")

    def test_no5(self):
        self._assert_components("Cno5", [0, 4], "C E")

    def test_add4(self):
        self._assert_components("Cadd4", [0, 4, 5, 7], "C E F G")

    def test_major_add4(self):
        # major add 4 is alternative notation for add4
        self._assert_components("CMadd4", [0, 4, 5, 7], "C E F G")

    def test_minor_add4(self):
        self._assert_components("Cmadd4", [0, 3, 5, 7], "C Eb F G")

    def test_sus4add9(self):
        self._assert_components("Csus4add9", [0, 5, 7, 14], "C F G D")

    def test_minor7_add11(self):
        self._assert_components('Cm7add11', [0, 3, 7, 10, 17], "C Eb G Bb F")

    def test_major7_add11(self):
        self._assert_components('CM7add11', [0, 4, 7, 11, 17], "C E G B F")

    def test_minormajor7_add11(self):
        self._assert_components('CmM7add11', [0, 3, 7, 11, 17], "C Eb G B F")

    def test_major7_add13(self):
        self._assert_components("CM7add13", [0, 4, 7, 9, 11, 14], "C E G A B D")


class TestChordComponentWithPitch(unittest.TestCase):

    def test_normal_chord(self):
        c = Chord("C")
        com = c.components_with_pitch(root_pitch=1)
        self.assertEqual(com, ["C1", "E1", "G1"])

    def test_minor_chord(self):
        c = Chord("Am")
        com = c.components_with_pitch(root_pitch=2)
        self.assertEqual(com, ["A2", "C3", "E3"])

    def test_slash_chord(self):
        c = Chord("Dm7/G")
        com = c.components_with_pitch(root_pitch=3)
        self.assertEqual(com, ["G3", "D4", "F4", "A4", "C5"])

    def test_add9_chord(self):
        c = Chord("Eadd9")
        com = c.components_with_pitch(root_pitch=5)
        self.assertEqual(com, ["E5", "G#5", "B5", "F#6"])

    def test_first_order_inversion(self):
        c = Chord("G/1")
        com = c.components_with_pitch(root_pitch=4)
        self.assertEqual(com, ["B4", "D5", "G5"])
        c2 = Chord("G13b9/1")
        com2 = c2.components_with_pitch(root_pitch=4)
        self.assertEqual(com2, ['B4', 'D5', 'F5', 'G#5', 'E6', 'G6'])

    def test_second_order_inversion(self):
        c = Chord("G/2")
        com = c.components_with_pitch(root_pitch=4)
        self.assertEqual(com, ["D5", "G5", "B5"])
        c2 = Chord("G13b9/2")
        com2 = c2.components_with_pitch(root_pitch=4)
        self.assertEqual(com2, ['D5', 'F5', 'G#5', 'E6', 'G6', 'B6'])

    def test_third_order_inversion(self):
        c = Chord("Cm7/3")
        com = c.components_with_pitch(root_pitch=4)
        self.assertEqual(com, ['Bb4', 'C5', 'Eb5', 'G5'])
        c2 = Chord("F#7/3")
        com2 = c2.components_with_pitch(root_pitch=4)
        self.assertEqual(com2, ['E5', 'F#5', 'A#5', 'C#6'])
        c3 = Chord("G13b9/3")
        com3 = c3.components_with_pitch(root_pitch=4)
        self.assertEqual(com3, ['F5', 'G#5', 'E6', 'G6', 'B6', 'D7'])

    def test_fourth_order_inversion(self):
        c = Chord("F7b9")
        com = c.components_with_pitch(root_pitch=4)
        self.assertEqual(com, ['F4', 'A4', 'C5', 'Eb5', 'Gb5'])
        c2 = Chord("G13b9/4")
        com2 = c2.components_with_pitch(root_pitch=4)
        self.assertEqual(com2, ['G#5', 'E6', 'G6', 'B6', 'D7', 'F7'])

    def test_fifth_order_inversion(self):
        c = Chord("G13b9/5")
        com = c.components_with_pitch(root_pitch=4)
        self.assertEqual(com, ['E6', 'G6', 'B6', 'D7', 'F7', 'G#7'])


if __name__ == '__main__':
    unittest.main()



import copy
from collections import OrderedDict
from typing import Tuple, List

from .constants.qualities import DEFAULT_QUALITIES
from .utils import note_to_val, val_to_note


class Quality:
    """ Chord quality """

    def __init__(self, name: str, components: Tuple[int, ...]):
        """ Constructor of chord quality

        :param name: name of quality
        :param components: components of quality
        """
        self._quality: str = name
        self.components: Tuple[int, ...] = components

    def __unicode__(self):
        return self._quality

    def __str__(self):
        return self._quality

    def __eq__(self, other):
        if not isinstance(other, Quality):
            raise TypeError(f"Cannot compare Quality object with {type(other)} object")
        return self.components == other.components

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def quality(self):
        """ Get name of quality """
        return self._quality

    def get_components(self, root='C', visible=False):
        """ Get components of chord quality

        :param str root: the root note of the chord
        :param bool visible: returns the name of notes if True
        :rtype: list[str|int]
        :return: components of chord quality
        """
        root_val = note_to_val(root)
        components = [v + root_val for v in self.components]

        if visible:
            components = [val_to_note(c, scale=root) for c in components]

        return components

    def append_on_chord(self, on_chord, root):
        """ Append on chord

        To create Am7/G
        q = Quality('m7')
        q.append_on_chord('G', root='A')

        :param str on_chord: bass note of the chord
        :param str root: root note of the chord
        """
        root_val = note_to_val(root)
        on_chord_val = note_to_val(on_chord) - root_val

        components = list(self.components)
        for idx, val in enumerate(self.components):
            if val % 12 == on_chord_val:
                components.remove(val)
                break

        if on_chord_val > root_val:
            on_chord_val -= 12

        if on_chord_val not in components:
            components = [on_chord_val] + [
                v for v in components if v % 12 != on_chord_val % 12
            ]

        self.components = tuple(components)


class QualityManager:
    """ Singleton class to manage the qualities """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(QualityManager, cls).__new__(cls)
            cls._instance.load_default_qualities()
        return cls._instance

    def load_default_qualities(self):
        self._qualities = OrderedDict([
            (q, Quality(q, c)) for q, c in DEFAULT_QUALITIES
        ])

    def get_quality(self, name: str, inversion: int = 0) -> Quality:
        if name not in self._qualities:
            raise ValueError(f"Unknown quality: {name}")
        # Create a new instance not to affect any existing instances
        q = copy.deepcopy(self._qualities[name])
        # apply requested inversion :
        for i in range(inversion):
            n = q.components[0]
            while n < q.components[-1]:
                n += 12
            q.components = q.components[1:] + (n,)
        return q

    def set_quality(self, name: str, components: Tuple[int, ...]):
        """ Set a Quality

        This method will not affect any existing Chord instances.
        :param name: name of quality
        :param components: components of quality
        """
        self._qualities[name] = Quality(name, components)

    def find_quality_from_components(self, components: List[int]):
        """ Find a quality from components

        :param components: components of quality
        """
        for q in self._qualities.values():
            if list(q.components) == components:
                return copy.deepcopy(q)
        return None



import unittest

from pychord import Chord


class TestChordCreations(unittest.TestCase):

    def test_normal_chord(self):
        c = Chord("C")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "")

    def test_invalid_chord(self):
        self.assertRaises(ValueError, Chord, "H")

    def test_minor_chord(self):
        c = Chord("Am")
        self.assertEqual(c.root, "A")
        self.assertEqual(c.quality.quality, "m")

    def test_minor_chord_with_minus(self):
        c = Chord("A-")
        self.assertEqual(c.root, "A")
        self.assertEqual(c.quality.quality, "-")

    def test_69_chord(self):
        c = Chord("C69")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "69")

    def test_m75_chord(self):
        c = Chord("Bm7-5")
        self.assertEqual(c.root, "B")
        self.assertEqual(c.quality.quality, "m7-5")

    def test_invalid_quality_chord(self):
        self.assertRaises(ValueError, Chord, "Csus3")

    def test_slash_chord(self):
        c = Chord("F/G")
        self.assertEqual(c.root, "F")
        self.assertEqual(c.quality.quality, "")
        self.assertEqual(c.on, "G")

    def test_minor_slash_chord(self):
        c = Chord("Dm/G")
        self.assertEqual(c.root, "D")
        self.assertEqual(c.quality.quality, "m")
        self.assertEqual(c.on, "G")

    def test_minor_7b5_chord(self):
        c = Chord("Dm7b5")
        self.assertEqual(c.root, "D")
        self.assertEqual(c.quality.quality, "m7b5")

    def test_invalid_slash_chord(self):
        self.assertRaises(ValueError, Chord, "C/H")

    def test_1st_order_inversion(self):
        c = Chord("C/1")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "")
        self.assertEqual(c.components(), ["E", "G", "C"])

    def test_2nd_order_inversion(self):
        c = Chord("C/2")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "")
        self.assertEqual(c.components(), ["G", "C", "E"])

    def test_inversion_complicated(self):
        c = Chord("Dm7b5/1")
        self.assertEqual(c.root, "D")
        self.assertEqual(c.quality.quality, "m7b5")
        self.assertEqual(c.components(), ["F", "G#", "C", "D"])

    def test_inversion_with_alternate_bass(self):
        c = Chord("C/1/F")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "")
        self.assertEqual(c.components(), ["F", "E", "G", "C"])

    def test_eq(self):
        c1 = Chord("C")
        c2 = Chord("C")
        self.assertEqual(c1, c2)

    def test_eq_alias(self):
        c1 = Chord("Cmaj7")
        c2 = Chord("CM7")
        self.assertEqual(c1, c2)

    def test_invalid_eq(self):
        c = Chord("C")
        with self.assertRaises(TypeError):
            print(c == 0)

    def test_components(self):
        c = Chord("C/E")
        quality_components_before = c.quality.components
        c.components()
        self.assertEqual(c.quality.components, quality_components_before)


class TestChordFromNoteIndex(unittest.TestCase):
    def test_note_1(self):
        chord = Chord.from_note_index(note=1, quality="", scale="Cmaj")
        self.assertEqual(chord, Chord("C"))

    def test_note_2(self):
        chord = Chord.from_note_index(note=2, quality="m7", scale="F#min")
        self.assertEqual(chord, Chord("G#m7"))

    def test_note_3(self):
        chord = Chord.from_note_index(note=3, quality="sus2", scale="Cmin")
        self.assertEqual(chord, Chord("Ebsus2"))

    def test_note_7(self):
        chord = Chord.from_note_index(note=7, quality="7", scale="Amin")
        self.assertEqual(chord, Chord("G7"))

    def test_note_8(self):
        chord = Chord.from_note_index(note=8, quality="", scale="Emaj")
        self.assertEqual(chord, Chord("E"))

    def test_note_0(self):
        with self.assertRaises(ValueError):
            Chord.from_note_index(note=0, quality="", scale="Cmaj")

    def test_note_9(self):
        with self.assertRaises(ValueError):
            Chord.from_note_index(note=9, quality="", scale="Fmaj")

    def test_diatonic_note_1(self):
        chord = Chord.from_note_index(note=1, quality="", diatonic=True, scale="Dmaj")
        self.assertEqual(chord, Chord("D"))

    def test_diatonic_note_2_mode(self):
        chord = Chord.from_note_index(note=2, quality="7", diatonic=True, scale="BLoc")
        self.assertEqual(chord, Chord("Cmaj7"))

    def test_diatonic_note_3_mode(self):
        chord = Chord.from_note_index(note=3, quality="m", diatonic=True, scale="G#Mix")
        self.assertEqual(chord, Chord("Cdim"))

    def test_diatonic_note_4_mode(self):
        chord = Chord.from_note_index(note=4, quality="-", diatonic=True, scale="AbDor")
        self.assertEqual(chord, Chord("C#"))

    def test_diatonic_note_nongeneric(self):
        with self.assertRaises(NotImplementedError):
            Chord.from_note_index(note=5, quality="sus", diatonic=True, scale="Fmaj")


if __name__ == '__main__':
    unittest.main()



from typing import Tuple, List
import re

from .quality import QualityManager, Quality
from .utils import NOTE_VAL_DICT

inversion_re = re.compile("/([0-9]+)")


def parse(chord: str) -> Tuple[str, Quality, List[str], str]:
    """Parse a string to get chord component

    :param chord: str expression of a chord
    :return: (root, quality, appended, on)
    """

    if len(chord) > 1 and chord[1] in ("b", "#"):
        root = chord[:2]
        rest = chord[2:]
    else:
        root = chord[:1]
        rest = chord[1:]

    def check_note(note: str):
        """ Raise ValueError if note is invalid """
        if note not in NOTE_VAL_DICT:
            raise ValueError(f"Invalid note {note}")

    check_note(root)

    inversion = 0
    inversion_m = inversion_re.search(rest)
    if inversion_m:
        inversion = int(inversion_m.group(1))
        rest = inversion_re.sub("", rest)

    on_chord_idx = rest.find("/")
    if on_chord_idx >= 0:
        on = rest[on_chord_idx + 1:]
        rest = rest[:on_chord_idx]
        check_note(on)
    else:
        on = ""
    quality = QualityManager().get_quality(rest, inversion)
    # TODO: Implement parser for appended notes
    appended: List[str] = []
    return root, quality, appended, on



# -*- coding: utf-8 -*-

import unittest

from pychord import QualityManager, Chord, find_chords_from_notes


class TestQuality(unittest.TestCase):
    def setUp(self):
        self.quality_manager = QualityManager()

    def test_eq(self):
        q1 = self.quality_manager.get_quality("m7-5")
        q2 = self.quality_manager.get_quality("m7-5")
        self.assertEqual(q1, q2)

    def test_eq_alias_maj9(self):
        q1 = self.quality_manager.get_quality("M9")
        q2 = self.quality_manager.get_quality("maj9")
        self.assertEqual(q1, q2)

    def test_eq_alias_m7b5(self):
        q1 = self.quality_manager.get_quality("m7-5")
        q2 = self.quality_manager.get_quality("m7b5")
        self.assertEqual(q1, q2)

    def test_eq_alias_min(self):
        q1 = self.quality_manager.get_quality("m")
        q2 = self.quality_manager.get_quality("min")
        q3 = self.quality_manager.get_quality("-")
        self.assertEqual(q1, q2)
        self.assertEqual(q1, q3)

    def test_invalid_eq(self):
        q = self.quality_manager.get_quality("m7")
        with self.assertRaises(TypeError):
            print(q == 0)


class TestQualityManager(unittest.TestCase):

    def test_singleton(self):
        quality_manager = QualityManager()
        quality_manager2 = QualityManager()
        self.assertIs(quality_manager, quality_manager2)


class TestOverwriteQuality(unittest.TestCase):
    def setUp(self):
        self.quality_manager = QualityManager()

    def tearDown(self):
        self.quality_manager.load_default_qualities()

    def test_overwrite(self):
        self.quality_manager.set_quality("11", (0, 4, 7, 10, 14, 17))
        chord = Chord("C11")
        self.assertEqual(chord.components(), ['C', 'E', 'G', 'Bb', 'D', 'F'])

    def test_find_from_components(self):
        self.quality_manager.set_quality("13", (0, 4, 7, 10, 14, 17, 21))
        chords = find_chords_from_notes(['C', 'E', 'G', 'Bb', 'D', 'F', 'A'])
        self.assertEqual(chords, [Chord("C13")])

    def test_keep_existing_chord(self):
        chord = Chord("C11")
        self.quality_manager.set_quality("11", (0, 4, 7, 10, 14, 17))
        self.assertEqual(chord.components(), ['C', 'G', 'Bb', 'D', 'F'])


if __name__ == '__main__':
    unittest.main()



from typing import List

from .constants import NOTE_VAL_DICT, SCALE_VAL_DICT


def note_to_val(note: str) -> int:
    """ Get index value of a note

    >>> note_to_val("C")
    0
    >>> note_to_val("B")
    11
    """
    if note not in NOTE_VAL_DICT:
        raise ValueError(f"Unknown note {note}")
    return NOTE_VAL_DICT[note]


def val_to_note(val: int, scale: str = "C") -> str:
    """ Return note by index in a scale

    >>> val_to_note(0)
    "C"
    >>> val_to_note(11, "D")
    "D#"
    """
    val %= 12
    return SCALE_VAL_DICT[scale][val]


def transpose_note(note: str, transpose: int, scale: str = "C") -> str:
    """ Transpose a note

    >>> transpose_note("C", 1)
    "Db"
    >>> transpose_note("D", 4, "A")
    "F#"
    """
    val = note_to_val(note)
    val += transpose
    return val_to_note(val, scale)


def display_appended(appended: List[str]) -> str:
    # TODO: Implement this
    return ""


def display_on(on_note: str) -> str:
    if on_note:
        return f"/{on_note}"
    return ""



import unittest

from pychord import Chord
from pychord.analyzer import get_all_rotated_notes, find_chords_from_notes, notes_to_positions


class TestNotesToPositions(unittest.TestCase):

    def test_one(self):
        pos = notes_to_positions(["C"], "C")
        self.assertEqual(pos, [0])

    def test_power(self):
        pos = notes_to_positions(["C", "G"], "C")
        self.assertEqual(pos, [0, 7])

    def test_major(self):
        pos = notes_to_positions(["D", "F#", "A"], "D")
        self.assertEqual(pos, [0, 4, 7])

    def test_seventh(self):
        pos = notes_to_positions(["E", "G#", "B", "D"], "E")
        self.assertEqual(pos, [0, 4, 7, 10])

    def test_add9(self):
        pos = notes_to_positions(["Ab", "C", "Eb", "Bb"], "Ab")
        self.assertEqual(pos, [0, 4, 7, 14])

    def test_major_add_9(self):
        # major add 9 is the same as add9
        self.test_add9()

    def test_ninth(self):
        pos = notes_to_positions(["F", "A", "C", "Eb", "G"], "F")
        self.assertEqual(pos, [0, 4, 7, 10, 14])

    def test_eleventh(self):
        pos = notes_to_positions(["G", "B", "D", "F", "A", "C"], "G")
        self.assertEqual(pos, [0, 4, 7, 10, 14, 17])

    def test_thirteenth(self):
        pos = notes_to_positions(["A", "C#", "E", "G", "B", "D", "F#"], "A")
        self.assertEqual(pos, [0, 4, 7, 10, 14, 17, 21])


class TestGetAllRotatedNotes(unittest.TestCase):

    def test_two(self):
        notes_list = get_all_rotated_notes(["C", "G"])
        self.assertEqual(notes_list, [["C", "G"], ["G", "C"]])

    def test_three(self):
        notes_list = get_all_rotated_notes(["C", "F", "G"])
        self.assertEqual(notes_list, [["C", "F", "G"], ["F", "G", "C"], ["G", "C", "F"]])


class TestFindChordsFromNotes(unittest.TestCase):

    def _assert_chords(self, notes, expected_chords):
        """ Validates that the specified notes translated to the expected chords.

        :param notes: The notes of the chord, either as a list of strings,
          e.g. ["G", "C", "D"] or a string, e.g. "G C D"
        :param expected_chords: the chords that the notes could translate to,
            specified as a list of strings, e.g. [ "Gsus4", "Csus2/G" ],
            or a single string if only one chord expected.
        """
        if isinstance(notes, str):
            notes = notes.split()
        c0 = find_chords_from_notes(notes)
        if isinstance(expected_chords, str):
            expected_chords = [expected_chords]
        self.assertEqual(c0, [Chord(c) for c in expected_chords])

    def test_major(self):
        chords = find_chords_from_notes(["C", "E", "G"])
        self.assertEqual(chords, [Chord("C")])

    def test_major_on_third(self):
        chords = find_chords_from_notes(["F#", "A", "D"])
        self.assertEqual(chords, [Chord("D/F#")])

    def test_major_on_fifth(self):
        chords = find_chords_from_notes(["B", "E", "G#"])
        self.assertEqual(chords, [Chord("E/B")])

    def test_dim(self):
        chords = find_chords_from_notes(["Eb", "Gb", "A"])
        self.assertEqual(chords, [Chord("Ebdim")])

    def test_sus4(self):
        chords = find_chords_from_notes(["G", "C", "D"])
        self.assertEqual(chords, [Chord("Gsus4"), Chord("Csus2/G")])

    def test_dim6(self):
        chords = find_chords_from_notes(["Eb", "Gb", "A", "C"])
        self.assertEqual(chords, [Chord("Ebdim7"), Chord("Gbdim7/Eb"), Chord("Adim7/Eb"), Chord("Cdim7/Eb")])

    def test_aug(self):
        chords = find_chords_from_notes(["F", "A", "Db"])
        self.assertEqual(chords, [Chord("Faug"), Chord("Aaug/F"), Chord("Dbaug/F")])

    def test_add9(self):
        chords = find_chords_from_notes(["C", "E", "G", "D"])
        self.assertEqual(chords, [Chord("Cadd9")])

    def test_m7b5(self):
        chords = find_chords_from_notes(["F#", "A", "C", "E"])
        self.assertEqual(chords, [Chord("F#m7-5"), Chord("Am6/F#"), Chord("C6b5/F#")])

    def test_m7dim5(self):
        chords = find_chords_from_notes(["F#", "A", "C", "E"])
        self.assertEqual(chords, [Chord("F#m7-5"), Chord("Am6/F#"), Chord("C6b5/F#")])

    def test_add4(self):
        chords = find_chords_from_notes(["C", "E", "F", "G"])
        self.assertEqual(chords, [Chord("Cadd4")])

    def test_minor_add4(self):
        chords = find_chords_from_notes(["C", "Eb", "F", "G"])
        self.assertEqual(chords, [Chord("Cmadd4")])

    def test_minor7_add11(self):
        self._assert_chords("C Eb G Bb F", ["Cm7add11", "F11/C"])

    def test_major7_add11(self):
        self._assert_chords("C E G B F", "CM7add11")

    def test_minormajor7_add11(self):
        self._assert_chords("C Eb G B F", "CmM7add11")

    def test_major7_add13(self):
        self._assert_chords("C E G A B D", "CM7add13")

    def test_idempotence(self):
        for _ in range(2):
            chords = find_chords_from_notes(["Eb", "Gb", "A", "C"])
            self.assertEqual(chords, [Chord("Ebdim7"), Chord("Gbdim7/Eb"), Chord("Adim7/Eb"), Chord("Cdim7/Eb")])
            self.assertEqual(chords[0].components(visible=True), ["Eb", "Gb", "A", "C"])



import sys

from setuptools import setup, find_packages
version = '1.2.0'

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 6)

if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("""PyChords only supports Python 3.6 and above.""")
    sys.exit(1)

setup(
    name='pychord',
    version=version,
    description="A library to handle musical chords in python.",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    keywords='music chord',
    author='Yuma Mihira',
    author_email='info@yuma.cloud',
    url='https://github.com/yuma-m/pychord',
    license='MIT',
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    zip_safe=True,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    test_suite='test',
    python_requires=">=3.6",
)
